allMegagunSounds = {}
bulletsInAir = {}
bulletTypesArr = {"7.62 mm", "20 mm", "Mini nuke"}
bulletTypes = {
	["7.62 mm"] = {
		["damage"] = 1,
		["spread_radius"] = 6,
		["time_to_hit_target"] = 1 / 30,
		["cooldown"] = 0,
		["barrel_speed"] = 85
	},
	["20 mm"] = {
		["damage"] = 3,
		["spread_radius"] = 6,
		["time_to_hit_target"] = 1 / 20,
		["cooldown"] = 3 / 60,
		["barrel_speed"] = 100
	},
	["Mini nuke"] = {
		["damage"] = 2.2,
		["spread_radius"] = 4,
		["constant_velocity"] = 100,
		["cooldown"] = 1 / 5,
		["explodes"] = true,
		["barrel_speed"] = 70
	}
}
ammoReserves = {
	["7.62 mm"] = {
		["numBelts"] = 100,
		["bulletsPerBelt"] = 3600,
	},
	["20 mm"] = {
		["numBelts"] = 80,
		["bulletsPerBelt"] = 1500
	},
	["Mini nuke"] = {
		["numBelts"] = 50,
		["bulletsPerBelt"] = 300
	}
}
activeAmmo = "7.62 mm"
gunMode = 0

-- 0 = inactive
-- 1 = winding up
-- 2 = winding down
-- 3 = actively firing
GUN_INACTIVE = 0
GUN_WINDING_UP = 1
GUN_WINDING_DOWN = 2
GUN_FIRING = 3

BULLET_FIRE_LOCATIONS = {
	Vec(17.3, 90, -13.5),
	Vec(17.3, 90, -29.5),
	Vec(24.6, 90, -25.6),
	Vec(10.5, 90, -25.6),
	Vec(24.6, 90, -17.6),
	Vec(10.5, 90, -17.6)
}

reloadTimer = 0
cooldownTimer = 0
recoilTimer = 0

bulletFireLocations = nil
barrelRotateVelocity = 0
barrelRotateAcceleration = 0

recoilStrength = nil
maxRecoil = nil
timeInterval = nil
recoilAmplitude = nil
recoilAmplitudeCoefficient = nil
function ResetRecoil()
	recoilStrength = 0
	maxRecoil = 0.2
	timeInterval = 0.5
	recoilAmplitude = maxRecoil
	recoilAmplitudeCoefficient = 1/4
	recoilTimer = 0
end

function CopyTable(orig)
    local orig_type = type(orig)
    local copy = nil

    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[CopyTable(orig_key)] = CopyTable(orig_value)
        end
        setmetatable(copy, CopyTable(getmetatable(orig)))
    else
        copy = orig
    end

    return copy
end

function MagicaVoxelPointToLocalPoint(vec)
	-- X Z -Y, VOX_SIZE = 0.05
	return VecScale(Vec(vec[1], vec[3], -vec[2]), 0.05)
end

function GetRandomSpreadValue(r)
	return ((2 * math.random()) - 1) * r
end

function VecPrint(vector)
	return "{" .. vector[1] .. ", " .. vector[2] .. ", " .. vector[3] .. "}"
end

function NumEqualsWithTolerance(val1, val2, tolerance)
	return math.abs(val1 - val2) <= tolerance
end

function VecEqualsWithTolerance(vec1, vec2, tolerance)
	return NumEqualsWithTolerance(vec1[1], vec2[1], tolerance) and NumEqualsWithTolerance(vec1[2], vec2[2], tolerance) and NumEqualsWithTolerance(vec1[3], vec2[3], tolerance)
end

function VecDist(vec1, vec2)
	return VecLength(VecSub(vec1, vec2))
end

function VecDist2D(vec1, vec2)
	vec1H = VecCopy(vec1)
	vec1H[2] = 0
	vec2H = VecCopy(vec2)
	vec2H[2] = 0
	return VecDist(vec1H, vec2H)
end

function RandomVector3D(length)
	local theta = math.random() * 2 * math.pi
	local a = math.random() * 2 - 1
	local b = math.sqrt(1 - math.pow(a, 2))
	return VecScale(Vec(b * math.cos(theta), b * math.sin(theta), a), length)
end

function PlayReloadSound()
	PlaySound(allMegagunSounds["reload"], GetPlayerTransform().pos, 1.0)
	reloadTimer = 0.95
end

function CalculateRecoil(direction, dt)
	recoilStrength = (maxRecoil - recoilAmplitude) + ((recoilAmplitude / 2) * ((direction * math.sin(((math.pi / timeInterval) * recoilTimer) - (math.pi / 2))) + 1))
	if direction > 0 then
		recoilStrength = (maxRecoil - recoilAmplitude) + ((recoilAmplitude / 2) * ((direction * math.sin(((math.pi / timeInterval) * recoilTimer) - (math.pi / 2))) + 1))
	else
		recoilStrength = 1.125 * recoilAmplitude * (math.cos((math.pi / timeInterval) * recoilTimer) - (1 / 9))
	end
	recoilAmplitude = recoilAmplitude * math.pow(recoilAmplitudeCoefficient, dt)
end

function LoadBeltIfNecessary(forceReload)
	forceReload = forceReload or false
	if GetBool("level.unlimitedammo") then return end

	-- Reload ammo	
	ammoReserves[activeAmmo]["currentBelt"] = math.max(ammoReserves[activeAmmo]["currentBelt"], 0)
	if forceReload or ammoReserves[activeAmmo]["currentBelt"] == 0 then
		local willPlaySound = false
		if ammoReserves[activeAmmo]["numBelts"] > 0 then
			willPlaySound = true
			ammoReserves[activeAmmo]["numBelts"] = ammoReserves[activeAmmo]["numBelts"] - 1
			ammoReserves[activeAmmo]["currentBelt"] = ammoReserves[activeAmmo]["bulletsPerBelt"]
		end

		if willPlaySound then
			PlayReloadSound()
		end
	end
end

function ShootMegagun()	
	local ct = GetCameraTransform()
	local forwardPoint = TransformToParentPoint(ct, Vec(0, 0, -1))
	local relativeDirection = VecNormalize(VecSub(forwardPoint, ct.pos))
	local relativeLength = VecDist(forwardPoint, ct.pos)
	local hit, dist = QueryRaycast(ct.pos, relativeDirection, relativeLength * 100)

	PlaySound(allMegagunSounds["fire"][math.random(1, #allMegagunSounds["fire"])], ct.pos, 0.55)

	-- Way more fun to have less spread at short range
	local localSpread = bulletTypes[activeAmmo]["spread_radius"] * (dist / 100)
	local numBullets = math.min(ammoReserves[activeAmmo]["currentBelt"], 6 * (GetTimeStep() * 60))

	BFLNum = 1
	for i = 1, numBullets do
		local thisBFL = bulletFireLocations[BFLNum]
		local bulletDirection = VecSub(VecAdd(ct.pos, VecScale(relativeDirection, dist)), thisBFL)
		bulletDirection[1] = bulletDirection[1] + GetRandomSpreadValue(localSpread)
		bulletDirection[2] = bulletDirection[2] + GetRandomSpreadValue(localSpread)
		bulletDirection[3] = bulletDirection[3] + GetRandomSpreadValue(localSpread)
		bulletDirection = VecNormalize(bulletDirection)
	
		local nextBullet = CopyTable(bulletTypes[activeAmmo])
		nextBullet.isInMotion = true
		nextBullet.isDead = false
		nextBullet.damage = nextBullet.damage * (math.random() + 0.5)
		nextBullet.age = 0
		nextBullet.pos = VecCopy(thisBFL)
		nextBullet.dir = VecCopy(bulletDirection)
		nextBullet.velocity = VecScale(nextBullet.dir, bulletTypes[activeAmmo]["constant_velocity"] or (dist / bulletTypes[activeAmmo]["time_to_hit_target"]))
		bulletsInAir[#bulletsInAir + 1] = nextBullet	

		BFLNum = BFLNum + 1
		if BFLNum > #bulletFireLocations then BFLNum = 1 end
	end

	ammoReserves[activeAmmo]["currentBelt"] = ammoReserves[activeAmmo]["currentBelt"] - numBullets
	LoadBeltIfNecessary()

	cooldownTimer = bulletTypes[activeAmmo]["cooldown"]
end

function ProjectileOperations(bullet)
	-- We need to check: does something intercept the path of our bullet this tick?
	local newPosition = VecAdd(bullet.pos, VecScale(bullet.velocity, GetTimeStep()))
	local relativeBullet = VecSub(newPosition, bullet.pos)
	local direction = VecNormalize(relativeBullet)
	local hit, dist = QueryRaycast(bullet.pos, direction, VecLength(relativeBullet))
	
	-- If so, then collide with it, otherwise just render the bullet's motion
	if hit then
		hitPos = VecAdd(bullet.pos, VecScale(direction, dist))
		bullet.isInMotion = false
		bullet.isDead = true

		if bullet.explodes then
			Explosion(hitPos, bullet.damage)
		else
			MakeHole(hitPos, bullet.damage, bullet.damage * 0.8, bullet.damage * 0.6)
		end

		ParticleReset()
		ParticleType("smoke")
		ParticleRadius(0.2)
		SpawnParticle(hitPos, Vec(0, math.random() + 0.2, 0), 1)
	else
		DrawLine(bullet.pos, newPosition)
	end

	-- We need to die if we're too old
	bullet.age = bullet.age + GetTimeStep()
	if bullet.age > 5 then
		bullet.isInMotion = false
		bullet.isDead = true
	end
	
	bullet.pos = newPosition
end

function init()
	RegisterTool("megagun", "Megagun", "MOD/vox/MEGAGUN.vox")
	SetBool("game.tool.megagun.enabled", true)

	ResetRecoil()

	-- Set default bullet number
	for k, v in pairs(ammoReserves) do
		ammoReserves[k]["currentBelt"] = ammoReserves[k]["bulletsPerBelt"]
	end

	allMegagunSounds["ambient"] = LoadLoop("MOD/snd/MEGAGUN_Firing_Ambient.ogg")
	allMegagunSounds["reload"] = LoadSound("MOD/snd/MEGAGUN_Reload.ogg")
	allMegagunSounds["winddown"] = LoadSound("MOD/snd/MEGAGUN_Spin_Down_Sound.ogg")
	allMegagunSounds["windup"] = LoadSound("MOD/snd/MEGAGUN_Wind_Up_Sound.ogg")

	allMegagunSounds["fire"] = {}
	for i = 1, 3 do
		allMegagunSounds["fire"][i] = LoadSound("MOD/snd/MEGAGUN_Firing_Sound_" .. i .. ".ogg")
	end
end

-- Updates once per frame
function tick(dt)
	-- Tick active bullets
	local cleaningUpBullets = {}

	for i = 1, #bulletsInAir do
		local thisBullet = bulletsInAir[i]
		if thisBullet == nil or thisBullet.isDead then
			cleaningUpBullets[#cleaningUpBullets + 1] = i
		elseif thisBullet.isInMotion then
			ProjectileOperations(thisBullet)
		end
	end

	for i = 1, #cleaningUpBullets do
		table.remove(bulletsInAir, cleaningUpBullets[i])
	end

	if GetBool("level.unlimitedammo") then
		ammoReserves[activeAmmo]["currentBelt"] = 99999
		ammoReserves[activeAmmo]["numBelts"] = 99999
	end

	SetInt("game.tool.megagun.ammo", ammoReserves[activeAmmo]["numBelts"])

	if GetString("game.player.tool") == "megagun" and GetBool("game.player.canusetool") and GetFloat("game.player.health") ~= 0 then
		-- Switch ammo type
		if InputPressed("c") and gunMode == GUN_INACTIVE then
			for i = 1, #bulletTypesArr do
				local k = bulletTypesArr[i]
				if k == activeAmmo then
					local newBulletIndex = i + 1
					if newBulletIndex > #bulletTypesArr then newBulletIndex = 1 end
					activeAmmo = bulletTypesArr[newBulletIndex]
					PlayReloadSound()
					break
				end
			end
		end
		
		-- Reload
		if InputPressed("r") and ammoReserves[activeAmmo]["currentBelt"] < ammoReserves[activeAmmo]["bulletsPerBelt"] then
			LoadBeltIfNecessary(true)
		end

		-- Time to wind up
		if ammoReserves[activeAmmo]["currentBelt"] > 0 and InputDown("usetool") and (gunMode == GUN_INACTIVE or gunMode == GUN_WINDING_DOWN) then
			PlaySound(allMegagunSounds["windup"], GetPlayerTransform().pos, 0.6)
			gunMode = GUN_WINDING_UP
		end

		-- Time to wind down
		if gunMode == GUN_WINDING_UP or gunMode == GUN_FIRING then
			if ammoReserves[activeAmmo]["currentBelt"] <= 0 or InputReleased("usetool") then
				PlaySound(allMegagunSounds["winddown"], GetPlayerTransform().pos, 0.6)
				gunMode = GUN_WINDING_DOWN
			end	
		end

		-- Recoil
		local toolBody = GetToolBody()
		if toolBody ~= 0 then
			SetToolTransform(Transform(Vec(0, 0, recoilStrength)))
		end
	else
		barrelRotateVelocity = 0
		barrelRotateAcceleration = 0
		gunMode = GUN_INACTIVE
	end
end

-- Updates always 60 times per second
undoingRecoil = false
function update(dt)
	if reloadTimer > 0 then
		reloadTimer = reloadTimer - dt
	end
	if cooldownTimer > 0 then
		cooldownTimer = cooldownTimer - dt
	end

	-- Undo recoil
	if gunMode == GUN_FIRING then
		if undoingRecoil then
			undoingRecoil = false
			ResetRecoil()
		end
		recoilTimer = recoilTimer + dt
	else
		if undoingRecoil or recoilStrength > 0 then
			if not undoingRecoil then
				undoingRecoil = true
				local initialStoppingRecoil = recoilStrength
				ResetRecoil()
				maxRecoil = initialStoppingRecoil
				recoilAmplitude = maxRecoil
				recoilAmplitudeCoefficient = 1/16
			end

			recoilTimer = recoilTimer + dt
			CalculateRecoil(-1, dt)
		end

		if undoingRecoil and recoilTimer >= 2.5 then
			undoingRecoil = false
			ResetRecoil()
		end
	end

	-- Winding up animation
	if gunMode == GUN_WINDING_UP then
		if barrelRotateAcceleration <= 0 then
			barrelRotateAcceleration = (bulletTypes[activeAmmo]["barrel_speed"] - barrelRotateVelocity) / 1.38
		end

		barrelRotateVelocity = barrelRotateVelocity + (barrelRotateAcceleration * dt)
		if barrelRotateVelocity >= bulletTypes[activeAmmo]["barrel_speed"] then
			barrelRotateVelocity = bulletTypes[activeAmmo]["barrel_speed"]
			barrelRotateAcceleration = 0
			gunMode = GUN_FIRING
		end
	end

	-- Winding down animation
	if gunMode == GUN_WINDING_DOWN then
		if barrelRotateAcceleration >= 0 then
			barrelRotateAcceleration = (0 - barrelRotateVelocity) / 1.39
		end

		barrelRotateVelocity = barrelRotateVelocity + (barrelRotateAcceleration * dt)
		if barrelRotateVelocity <= 0 then
			barrelRotateVelocity = 0
			barrelRotateAcceleration = 0
			gunMode = GUN_INACTIVE
		end	
	end

	local toolBody = GetToolBody()
	if GetString("game.player.tool") == "megagun" and toolBody ~= 0 then
		-- Get some relative positions based off of the tool body
		local toolTrans = GetBodyTransform(toolBody)
		bulletFireLocations = {}
		for i = 1, #BULLET_FIRE_LOCATIONS do
			bulletFireLocations[#bulletFireLocations + 1] = TransformToParentPoint(toolTrans, MagicaVoxelPointToLocalPoint(BULLET_FIRE_LOCATIONS[i]))
		end

		-- Physically rotate the barrel
		local megagunShapes = GetBodyShapes(GetToolBody())
		local barrelShape = megagunShapes[2]
		local attachmentPoint = Transform(MagicaVoxelPointToLocalPoint(Vec(17.5, 0, -21.2)))
		local barrelTrans = TransformToLocalTransform(attachmentPoint, GetShapeLocalTransform(barrelShape))
		attachmentPoint.rot = QuatEuler(0, 0, barrelRotateVelocity)
		SetShapeLocalTransform(barrelShape, TransformToParentTransform(attachmentPoint, barrelTrans))

		-- Firing effects
		if gunMode == GUN_FIRING then
			PlayLoop(allMegagunSounds["ambient"], GetPlayerTransform().pos, 0.8, false)
			CalculateRecoil(1, dt)
		end

		-- Firing bullets
		if gunMode == GUN_FIRING and reloadTimer <= 0 and cooldownTimer <= 0 then
			ShootMegagun()
		end
	end
end


-- Configuration UI
function draw()
	if GetString("game.player.tool") == "megagun" and GetBool("game.player.interactive") then
		-- Ammo indicator
		if not GetBool("level.unlimitedammo") then
			UiPush()
				UiTranslate(UiCenter(), UiHeight()-90)
				UiAlign("center middle")
				UiColor(1, 1, 1)
				UiFont("bold.ttf", 24)
				UiTextOutline(0, 0, 0, 1, 0.1)
				UiText(string.format("%.0f", ammoReserves[activeAmmo]["currentBelt"]) .. " / " .. string.format("%.0f", ammoReserves[activeAmmo]["bulletsPerBelt"]))
			UiPop()
		end

		-- Ammo type
		UiPush()
			UiTranslate(70, UiHeight() - (30 * #bulletTypesArr + 20 + 80))
			UiAlign("bottom left")
			UiColor(1, 1, 1)
			UiFont("regular.ttf", 24)
			UiTextOutline(0, 0, 0, 1, 0.1)
			UiText("Press C to switch ammo type")
			UiTranslate(0, 20)
			for i = 1, #bulletTypesArr do
				UiColor(0.5, 0.5, 0.5)
				if activeAmmo == bulletTypesArr[i] then UiColor(1, 1, 1) end
				UiTranslate(0, 30)
				UiText(bulletTypesArr[i])
			end
		UiPop()
	end
end